package ru.dbotthepony.mc.otm.capability.drive;

import net.minecraft.CrashReport;
import net.minecraft.ReportedException;
import net.minecraft.Util;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.NbtIo;
import net.minecraft.server.MinecraftServer;
import net.minecraftforge.event.TickEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fmlserverevents.FMLServerStartedEvent;
import net.minecraftforge.fmlserverevents.FMLServerStoppingEvent;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;

import javax.annotation.Nullable;
import java.io.File;
import java.lang.reflect.Array;
import java.security.Provider;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.locks.LockSupport;
import java.util.function.Function;
import java.util.function.Supplier;


/**
 * Let me answer everyone's questions about:
 *
 * Why?
 *
 * There are several reasons:
 * 0. This data can get very large very quickly, even when playing singleplayer (and much quicker and bigger when hosting a dedicated server)
 * 1. This data can not be stored inside ItemStack.ForgeCaps due to it's size
 * 2. This data can not be stored inside unshared (server only) nbt tag because, again, mods prone to use and interact with
 *    it wrong, causing loss of stored data or mods exposing full content of a drive inside their own tag (which cause very real "NBT size too large"
 *    network kicks, often locking players out of server
 * 3. net.minecraft.world.level.saveddata.SaveData is for storing everything inside one dat file, which
 *    is performance tanking, because we have to write *entire* NBT on each save, not the data of Drive which got changed
 */
public class DrivePool {
	private static final Logger LOGGER = LogManager.getLogger();

	public static class Entry {
		public final UUID id;
		public final IMatteryDrive drive;
		public long last_access = System.currentTimeMillis();
		public long soft_sync_on = last_access + 8_000; // soft sync in 4 seconds
		public long hard_sync_on = last_access + 30_000; // hard sync in 4 seconds

		Entry(UUID id, IMatteryDrive drive) {
			this.id = id;
			this.drive = drive;
		}

		@Nullable
		public CompoundTag sync(boolean force) {
			if (!drive.isDirty()) {
				soft_sync_on = System.currentTimeMillis() + 8_000;
				hard_sync_on = System.currentTimeMillis() + 30_000;
				next_save = Math.min(next_save, soft_sync_on);
				return null;
			}

			if (force || hard_sync_on <= System.currentTimeMillis() || soft_sync_on <= System.currentTimeMillis()) {
				var tag = drive.serializeNBT();
				drive.markClean();
				soft_sync_on = System.currentTimeMillis() + 8_000;
				hard_sync_on = System.currentTimeMillis() + 30_000;
				next_save = Math.min(next_save, soft_sync_on);
				last_access = System.currentTimeMillis();
				return tag;
			}

			return null;
		}

		public IMatteryDrive access() {
			last_access = System.currentTimeMillis();
			soft_sync_on = last_access + 8_000;
			next_save = Math.min(hard_sync_on, Math.min(next_save, soft_sync_on));
			return drive;
		}
	}

	private static HashMap<UUID, Entry> pool = new HashMap<>();
	private static Thread thread;
	private static boolean stopping = false;
	private static ReportedException exception;
	public static final String DATA_PATH = "data/otm_drives";

	private static long next_save = Long.MAX_VALUE;

	public static <T extends IMatteryDrive> T get(UUID id, Function<CompoundTag, T> factory_stored, Supplier<T> factory_empty) {
		var get = pool.get(id);

		if (get != null) {
			return (T) get.access();
		}

		var file = new File(base_directory, id.toString() + ".dat");

		if (file.exists()) {
			try {
				var factorize = factory_stored.apply(NbtIo.readCompressed(file));
				pool.put(id, new Entry(id, factorize));
				return factorize;
			} catch(Throwable error) {
				var report = new CrashReport("Reading OTM Drive from disk", error);
				var category = report.addCategory("Corrupt drive");
				category.setDetail("UUID", id);
				throw new ReportedException(report);
			}
		}

		var factorize = factory_empty.get();
		pool.put(id, new Entry(id, factorize));

		return factorize;
	}

	public static void put(UUID id, IMatteryDrive drive) {
		pool.put(id, new Entry(id, drive));
	}

	private record BacklogLine(UUID file, CompoundTag tag, IMatteryDrive drive) {
		@Override
		public boolean equals(Object obj) {
			return obj instanceof BacklogLine line && line.file.equals(file);
		}
	}

	private static List<BacklogLine> backlog = new ArrayList<>();
	private static File base_directory;

	private static void thread(MinecraftServer server) {
		while (true) {
			LockSupport.park();

			if (stopping) {
				return;
			}

			try {
				sync();
			} catch (ReportedException error) {
				exception = error;
				return;
			}
		}
	}

	@SubscribeEvent
	public static void onServerPostTick(TickEvent.ServerTickEvent event) {
		if (event.phase == TickEvent.Phase.END) {
			if (exception != null) {
				throw exception;
			}

			//if (next_save < System.currentTimeMillis()) {
				writeBacklog(true);
			//}
		}
	}

	@SubscribeEvent
	public static void serverStartEvent(FMLServerStartedEvent event) {
		if (thread != null && thread.isAlive()) {
			LOGGER.error("Attempting to start another DrivePool I/O thread when already running one!");
			return;
		}

		pool.clear();

		final var server = event.getServer();
		base_directory = new File(server.storageSource.getWorldDir().toFile(), DATA_PATH);
		base_directory.mkdirs();

		thread = new Thread(null, () -> DrivePool.thread(server), "Overdrive That Matters DrivePool I/O");
		thread.start();
	}

	@SubscribeEvent
	public static void serverStopEvent(FMLServerStoppingEvent event) {
		if (thread != null && thread.isAlive()) {
			stopping = true;
			LockSupport.unpark(thread);
		}

		writeBacklog(true);
		sync();

		pool.clear();
	}

	public static void writeBacklog(boolean force) {
		var sync_required = false;
		ArrayList<UUID> remove_keys = null;

		for (var entry : pool.entrySet()) {
			var tag = entry.getValue().sync(force);

			if (tag != null) {
				LOGGER.info("Syncing {}", entry.getKey());

				var index = backlog.indexOf(entry.getKey());

				if (index != -1) {
					backlog.set(index, new BacklogLine(entry.getKey(), tag, entry.getValue().drive));
				} else {
					backlog.add(new BacklogLine(entry.getKey(), tag, entry.getValue().drive));
				}

				sync_required = true;
			} else if (entry.getValue().last_access < System.currentTimeMillis() - 600_000) {
				// no access or changes in 600 seconds - unload
				if (remove_keys == null)
					remove_keys = new ArrayList<>();

				remove_keys.add(entry.getKey());
			}
		}

		if (remove_keys != null) {
			for (var key : remove_keys) {
				pool.remove(key);
			}
		}

		if (sync_required) {
			LockSupport.unpark(thread);
		}

		if (pool.size() == 0) {
			next_save = Long.MAX_VALUE;
		}
	}

	public static void sync() {
		var copy = backlog;
		backlog = new ArrayList<>();

		for (var entry : copy) {
			try {
				var file = new File(base_directory, entry.file.toString() + ".dat");
				LOGGER.info("Writing {}", entry.file);
				NbtIo.writeCompressed(entry.tag, file);
			} catch(Throwable error) {
				var report = new CrashReport("Syncing OTM Drives state to disk", error);
				var category = report.addCategory("Corrupt drive");
				category.setDetail("UUID", entry.file.toString());
				category.setDetail("Stored item count", entry.drive.getStoredCount());
				category.setDetail("Capacity", entry.drive.getCapacity());

				if (entry.drive instanceof MatteryDrive drive) {
					category.setDetail("Amount of different stacks", drive.different_stacks);
					category.setDetail("Max amount of different stacks", drive.max_different_stacks);
				}

				throw new ReportedException(report);
			}
		}
	}
}
